# 常见约束

- 用于限制表中的数据，为了保证表中的数据的准确和可靠性

## 约束类型

- NOT NULL：非空，用于保证该字段的值不能为空
- DEFAULT：默认，用于保证该字段有默认值
- PRIMARY KEY：主键，用于保证该字段的值具有唯一性，并且非空
- UNIQUE：唯一，用于保证该字段的值具有唯一性，可以为空，为空的时候可以存在多个 NULL
- CHECK：检查约束，mysql 中不支持
- FOREIGN KEY：外键，用于限制两个表的关系，用于保证该字段的值必须来自于主表的关联列的值，在从表添加外键约束，用于引用主表中某列的值
  - 要求在从表设置外键关系
  - 从表的外键列的类型和主表的关联列的类型要求一致或兼容，名称无要求
  - 主表的关联列必须是一个key（一般是主键或唯一）
  - 插入数据时，先插入主表，再插入从表，删除数据时，先删除从表，再删除主表

## 分类

- 列级约束：六大约束语法上都支持，但外键约束没有效果
- 表级约束：除了非空、默认，其他的都支持

## 创建表时添加约束

### 添加列级约束

```mysql
CREATE TABLE stuinfo(
    # 主键约束
	id INT PRIMARY KEY,
    # 非空约束
	stuName VARCHAR(20) NOT NULL UNIQUE,
    # 检查约束（MySQL不支持，语法不报错）
	gender CHAR(1) CHECK(gender='男' OR gender ='女'),
    # 唯一约束
	seat INT UNIQUE,
    # 默认约束
	age INT DEFAULT 18,
    # 外键约束 
	majorId INT REFERENCES major(id)
);
```

### 添加表级约束

```mysql
CREATE TABLE stuinfo(
	id INT,
	stuname VARCHAR(20),
	gender CHAR(1),
	seat INT,
	age INT,
	majorid INT,
	
    # 主键约束
	CONSTRAINT pk PRIMARY KEY(id),
    # 唯一约束
	CONSTRAINT uq UNIQUE(seat),
    # 检查约束（MySQL不支持，语法不报错）
	CONSTRAINT ck CHECK(gender ='男' OR gender = '女'),
    # 外键约束 
	CONSTRAINT fk_stuinfo_major FOREIGN KEY(majorid) REFERENCES major(id)
);

```

## 修改表时添加约束

### 添加列级约束

#### 添加非空约束

```mysql
ALTER TABLE stuinfo MODIFY COLUMN stuname VARCHAR(20) NOT NULL;
```

#### 添加默认约束

```mysql
ALTER TABLE stuinfo MODIFY COLUMN age INT DEFAULT 18;
```

#### 添加主键约束

```mysql
# 列级约束
ALTER TABLE stuinfo MODIFY COLUMN id INT PRIMARY KEY;
# 表级约束
ALTER TABLE stuinfo ADD PRIMARY KEY(id);
```

#### 添加唯一约束

```mysql
# 列级约束
ALTER TABLE stuinfo MODIFY COLUMN seat INT UNIQUE;
# 表级约束
ALTER TABLE stuinfo ADD UNIQUE(seat);
```

#### 添加外键约束

```mysql
ALTER TABLE stuinfo ADD CONSTRAINT fk_stuinfo_major FOREIGN KEY(majorid) REFERENCES major(id); 
```

## 修改表时删除约束

### 删除非空约束

```mysql
ALTER TABLE stuinfo MODIFY COLUMN stuname VARCHAR(20) NULL;
```

### 删除默认约束

```mysql
ALTER TABLE stuinfo MODIFY COLUMN age INT;
```

### 删除主键约束

```mysql
ALTER TABLE stuinfo DROP PRIMARY KEY;
```

### 删除唯一约束

```mysql
ALTER TABLE stuinfo DROP INDEX seat;
```

### 删除外键

```mysql
ALTER TABLE stuinfo DROP FOREIGN KEY fk_stuinfo_major;
```

