package com.github.zj.dreamly.datastructure.tree.segmenttree;

/**
 * <h2>segmenttree</h2>
 *
 * @author: 苍海之南
 * @since: 2019-09-29 08:51
 **/
public class SegmentTree<E> {

	private E[] tree;
	private E[] data;
	private Merger<E> merger;

	@SuppressWarnings("unchecked")
	public SegmentTree(E[] arr, Merger<E> merger) {

		this.merger = merger;

		data = (E[]) new Object[arr.length];
		for (int i = 0; i < arr.length; i++) {
			data[i] = arr[i];
		}

		tree = (E[]) new Object[4 * arr.length];
		buildSegmentTree(0, 0, arr.length - 1);
	}

	/**
	 * 在treeIndex的位置创建表示区间[l...r]的线段树
	 */
	private void buildSegmentTree(int treeIndex, int l, int r) {

		if (l == r) {
			tree[treeIndex] = data[l];
			return;
		}

		int leftTreeIndex = leftChild(treeIndex);
		int rightTreeIndex = rightChild(treeIndex);

		// int mid = (l + r) / 2;
		int mid = l + (r - l) / 2;
		buildSegmentTree(leftTreeIndex, l, mid);
		buildSegmentTree(rightTreeIndex, mid + 1, r);

		tree[treeIndex] = merger.merge(tree[leftTreeIndex], tree[rightTreeIndex]);
	}

	public int getSize() {
		return data.length;
	}

	public E get(int index) {
		if (index < 0 || index >= data.length) {
			throw new IllegalArgumentException("Index is illegal.");
		}
		return data[index];
	}

	/**
	 * 返回完全二叉树的数组表示中，一个索引所表示的元素的左孩子节点的索引
	 */
	private int leftChild(int index) {
		return 2 * index + 1;
	}

	/**
	 * 返回完全二叉树的数组表示中，一个索引所表示的元素的右孩子节点的索引
	 */
	private int rightChild(int index) {
		return 2 * index + 2;
	}

	/**
	 * 返回区间[queryL, queryR]的值
	 */
	public E query(int queryLeft, int queryRight) {

		if (queryLeft < 0 || queryLeft >= data.length ||
			queryRight < 0 || queryRight >= data.length || queryLeft > queryRight) {
			throw new IllegalArgumentException("Index is illegal.");
		}

		return query(0, 0, data.length - 1, queryLeft, queryRight);
	}

	/**
	 * 在以treeIndex为根的线段树中[l...r]的范围里，搜索区间[queryL...queryR]的值
	 */
	private E query(int treeIndex, int l, int r, int queryLeft, int queryRight) {

		if (l == queryLeft && r == queryRight) {
			return tree[treeIndex];
		}

		int mid = l + (r - l) / 2;
		// treeIndex的节点分为[l...mid]和[mid+1...r]两部分

		int leftTreeIndex = leftChild(treeIndex);
		int rightTreeIndex = rightChild(treeIndex);
		if (queryLeft >= mid + 1) {
			return query(rightTreeIndex, mid + 1, r, queryLeft, queryRight);
		} else if (queryRight <= mid) {
			return query(leftTreeIndex, l, mid, queryLeft, queryRight);
		}

		E leftResult = query(leftTreeIndex, l, mid, queryLeft, mid);
		E rightResult = query(rightTreeIndex, mid + 1, r, mid + 1, queryRight);
		return merger.merge(leftResult, rightResult);
	}

	/**
	 * 将index位置的值，更新为e
	 */
	public void set(int index, E e) {

		if (index < 0 || index >= data.length) {
			throw new IllegalArgumentException("Index is illegal");
		}

		data[index] = e;
		set(0, 0, data.length - 1, index, e);
	}

	/**
	 * 在以treeIndex为根的线段树中更新index的值为e
	 */
	private void set(int treeIndex, int l, int r, int index, E e) {

		if (l == r) {
			tree[treeIndex] = e;
			return;
		}

		int mid = l + (r - l) / 2;

		// treeIndex的节点分为[l...mid]和[mid+1...r]两部分
		int leftTreeIndex = leftChild(treeIndex);
		int rightTreeIndex = rightChild(treeIndex);
		if (index >= mid + 1) {
			set(rightTreeIndex, mid + 1, r, index, e);
		} else // index <= mid
		{
			set(leftTreeIndex, l, mid, index, e);
		}

		tree[treeIndex] = merger.merge(tree[leftTreeIndex], tree[rightTreeIndex]);
	}

	@Override
	public String toString() {
		StringBuilder res = new StringBuilder();
		res.append('[');
		for (int i = 0; i < tree.length; i++) {
			if (tree[i] != null) {
				res.append(tree[i]);
			} else {
				res.append("null");
			}

			if (i != tree.length - 1) {
				res.append(", ");
			}
		}
		res.append(']');
		return res.toString();
	}
}
