### 简单工厂模式

- 定义：由一个工厂对象决定创建出来哪一种产品类的实例
- 类型：创建型，但不属于GOF23种设计模式

#### 适用场景

- 工厂类负责创建的对象比较少
- 客户端（应用层）只知道传入工厂类的参数，对于如何创建对象（逻辑）不关心

#### 优点

- 只需要传入一个正确的参数，就可以获取你所需要的对象而无需知道其创建细节

#### 缺点

- 工厂类的职责相对较重，增加新的产品，都需要修改工厂类的判断逻辑，违背开闭原则

#### 源码解读

- `calendar`
- `com,mysql.jdbc.Driver`
- `LoggerFactory`

### 工厂方法模式

- 定义：定义一个创建对象的接口，但让实现这个接口的类来决定实例化哪个类，工厂方法让类的实例化推迟到子类种进行
- 类型：创建型

#### 适用场景

- 创建对象需要大量重复代码
- 客户端（应用层）不依赖产品类实例如何被创建，实现等细节
- 一个类通过其子类来指定创建哪个对象

#### 优点

- 用户只需要关心所需产品对应的工厂，无需关心创建细节
- 加入新产品符合开闭原则，提高可扩展性

#### 缺点

- 类的个数容易过多，增加复杂度
- 增加了系统的抽象性和理解难度

#### 源码解读

- `collection`
- `URLStreamHandlerFactory`

### 抽象工厂模式

- 定义：抽象工厂模式提供一个创建一系列相关接口或相互依赖对象的接口
- 无需指定它们具体的类
- 类型：创建型

#### 适用场景

- 客户端（应用层）不依赖于产品类实例如何被创建，实现等细节
- 强调一系列相关的产品对象（属于同一产品族）一起使用，创建对象需要大量重复代码
- 提供一个产品类的库，所有的产品以同样的接口出现，从而使客户端不依赖具体实现

#### 优点

- 具体产品在应用层代码隔离，无需关心创建细节
- 将一个系列的产品族统一到一起创建

#### 缺点

- 规定了所有可能被创建的产品集合，产品族种扩展新的产品困难，需要修改抽象工厂的接口
- 增加了系统的抽象性和理解难度

#### 源码解读

- `java.sql.connection`

#### 产品等级结构和产品族

### 建造者模式

- 定义：将一个复杂对象的构建和它的表示分离，使得同样的构建过程可以创建不同的表示
- 用户只需要指定需要建造的类型就可以得到它们，建造过程及细节不需要知道
- 类型：创建型

#### 适用场景

- 如果一个对象有非常复杂的内部结构（很多属性）
- 想把复杂对象的创建和使用分离

#### 优点

- 封装性好，创建和使用分离
- 扩展性好，建造类之间独立，一定程度上解耦

#### 缺点

- 产生多余的Builder对象
- 产品内部发生变化，建造者都要修改，成本较大

#### 源码解读

- `StringBuilder`
- `ImmutableSet` 
- `SqlSessionFactoryBuilder`

### 单例模式

- 定义：保证一个类仅有一个实例，并提供一个全局访问点
- 类型：创建型

#### 适用场景

- 想确保任何情况下都绝对只有一个实例

#### 优点

- 在内存里只有一个实例，减少了内存开销
- 可以避免对资源的多重占用
- 设置全局访问点，严格控制访问

#### 缺点

- 没有接口，扩展困难

#### 重点

- 私有构造器
- 线程安全
- 延迟加载
- 序列化和反序列化安全
- 反射

#### 实用技能

- 反编译
- 内存原理
- 多线程Debug（右键小红点（idea），frames切换线程）

#### 单例-相关设计模式

- 单例模式和工厂模式
- 单例模式和享元模式

#### 【懒汉式】DoubleCheck实现单例

```java
public class LazyDoubleCheckSingleton {
    private volatile static LazyDoubleCheckSingleton lazyDoubleCheckSingleton = null;
    private LazyDoubleCheckSingleton(){

    }
    public static LazyDoubleCheckSingleton getInstance(){
        if(lazyDoubleCheckSingleton == null){
            synchronized (LazyDoubleCheckSingleton.class){
                if(lazyDoubleCheckSingleton == null){
                    lazyDoubleCheckSingleton = new LazyDoubleCheckSingleton();
                }
            }
        }
        return lazyDoubleCheckSingleton;
    }
}
```

#### 【懒汉式】基于类初始化延迟加载解决方案（静态内部类）实现单例

```java
public class StaticInnerClassSingleton {
    private static class InnerClass{
        private static StaticInnerClassSingleton staticInnerClassSingleton = new StaticInnerClassSingleton();
    }
    public static StaticInnerClassSingleton getInstance(){
        return InnerClass.staticInnerClassSingleton;
    }
    /**
     * 解决反射攻击
     */
    private StaticInnerClassSingleton(){
        if(InnerClass.staticInnerClassSingleton != null){
            throw new RuntimeException("单例构造器禁止反射调用");
        }
    }
}
```

#### 【饿汉式】实现单例

```java
public class HungrySingleton implements Serializable,Cloneable{

    private final static HungrySingleton hungrySingleton;

    static{
        hungrySingleton = new HungrySingleton();
    }
    /**
     * 解决反射攻击
     */
    private HungrySingleton(){
        if(hungrySingleton != null){
            throw new RuntimeException("单例构造器禁止反射调用");
        }
    }
    public static HungrySingleton getInstance(){
        return hungrySingleton;
    }

    /**
     * 解决因序列化导致破坏单例模式问题
     */
    private Object readResolve(){
        return hungrySingleton;
    }

    @Override
    protected Object clone() throws CloneNotSupportedException {
        return getInstance();
    }
}
```

#### 枚举实现单例

```java
public enum EnumInstance {
    INSTANCE{
        @Override
        protected  void printTest(){
            System.out.println("Geely Print Test");
        }
    };
    protected abstract void printTest();
    private Object data;

    public Object getData() {
        return data;
    }

    public void setData(Object data) {
        this.data = data;
    }
    public static EnumInstance getInstance(){
        return INSTANCE;
    }

}
```

#### JAD反编译工具

<https://varaneckas.com/jad/>

#### 容器单例（有线程安全隐患）

```java
public class ContainerSingleton {

    private ContainerSingleton(){

    }
    private static Map<String,Object> singletonMap = new HashMap<String,Object>();

    public static void putInstance(String key,Object instance){
        if(StringUtils.isNotBlank(key) && instance != null){
            if(!singletonMap.containsKey(key)){
                singletonMap.put(key,instance);
            }
        }
    }

    public static Object getInstance(String key){
        return singletonMap.get(key);
    }
}
```

#### ThreadLocal线程单例

```java
public class ThreadLocalInstance {
    private static final ThreadLocal<ThreadLocalInstance> threadLocalInstanceThreadLocal
             = new ThreadLocal<ThreadLocalInstance>(){
        @Override
        protected ThreadLocalInstance initialValue() {
            return new ThreadLocalInstance();
        }
    };
    private ThreadLocalInstance(){

    }

    public static ThreadLocalInstance getInstance(){
        return threadLocalInstanceThreadLocal.get();
    }
}
```

#### 源码解读

- Runtime
- AbstractFactoryBean
- ErrorContext

### 原型模式

- 定义：指原型实例指定创建对象的种类，并且通过拷贝这些原型创建新的对象
- 不需要知道任何创建细节，不调用构造函数
- 类型：创建型

#### 适用场景

- 类初始化消耗较多资源
- new产生的一个对象需要非常繁琐的过程，数据准备，访问权限等
- 构造函数比较复杂
- 循环体中生产大量对象时

#### 优点

- 原型模式性能比直接new一个对象性能高
- 简化创建过程

#### 缺点

- 必须配置克隆方法
- 对克隆复杂对象或对克隆出的对象进行复杂改造时，容易引入风险
- 深拷贝，浅拷贝要运用得当

#### 扩展

- 深克隆
- 浅克隆

#### 实现

1. 实现`Cloneable`
2. 重写`clone`方法（默认父类实现为浅克隆）

#### 源码解读

- `ArrayList`
- `HashMap`
- `Object`
- `cloneable`
- `CacheKey`